
//		Primes.cpp - implements @ISPRIME, @PRIME, @LPRIME, and related functions


#define UNICODE							1
#define _UNICODE						1

#ifdef PRIMES_RAND_UI64
#define _CRT_RAND_S
#endif	//	PRIMES_RAND_UI64

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR						0
#define VER_MINOR						99
#define VER_BUILD						16
#define VER_PATCH						0

#define PLUGIN_NAME						L"Primes"

#ifdef _WIN64
#define PLATFORM						L"x64"
#else
#define PLATFORM						L"x86"
#endif

#define MAX_ARG_LEN						8192
#define SHORT_BUF_LEN					1024

#define ERRHANDLE						GetStdHandle( STD_ERROR_HANDLE )


#define PRIMES_DECIMAL					0
#define PRIMES_HEX						1
#define PRIMES_HEX_ALT					2



//		The maximum number of arguments for @GCF and @LCM.  This is
//		a completely arbitrary number, of course.

#define MAX_VALUES						32

//		The maximum number of _different_ prime factors which a value
//		in the unsigned 64-bit range can have.

#define MAX_UNIQUE_PRIME_FACTORS		15


//		The largest prime in the 64-bit range :

#define LARGEST_PRIME					0xFFFFFFFFFFFFFFC5



#ifdef PRIMES_MINIMAL

const wchar_t PluginFeatures[] = L"@isprime,@prime";

#elif defined PRIMES_RAND_UI64

const wchar_t PluginFeatures[] = L"@factors,@gcf,@isprime,@isqrt,@lcm,@lprime,@prime,_randui64";

#else	//	PRIMES_MINIMAL

const wchar_t PluginFeatures[] = L"@factors,@gcf,@isprime,@isqrt,@lcm,@lprime,@prime";

#endif	//	PRIMES_MINIMAL



const wchar_t F_IsPrimeHelpText[]	= L"@ISPRIME[n] :  Primeness test  \r\n\
\n\
    n\ta 64-bit unsigned integer  \n\
\n\
Returns 1 if n has exactly two factors,  \n\
0 otherwise.  (By this definition, zero  \n\
and one are not prime.) \n";


const wchar_t F_PrimeHelpText[]		= L"@PRIME[n] :  Rightward prime  \r\n\
\n\
    n\ta 64-bit unsigned integer  \n\
\n\
Returns the smallest prime number not less  \n\
than n.\n";


#ifndef PRIMES_MINIMAL

const wchar_t F_FactorsHelpText[]	= L"@FACTORS[n,m,p] :  Prime factors  \r\n\
\n\
    n\ta 64-bit unsigned integer\n\
    m\tmultiplication symbol ( \u00d7 )  \n\
    p\tpower symbol (not used)\n\
\n\
Returns a list of prime factors of n.\n";


const wchar_t F_GcfHelpText[]		= L"@GCF[n1,n2,...] :  Greatest common factor  \r\n\
\n\
    n1, n2, ...\t64-bit unsigned integers  \n\
\n\
A maximum of 32 values are allowed.\n";


const wchar_t F_IsqrtHelpText[]		= L"@ISQRT[n] :  Integer square root  \r\n\
\n\
    n\ta 64-bit unsigned integer  \n\
\n\
Returns the integer square root of n.  \n";


const wchar_t F_LcmHelpText[]		= L"@LCM[n1,n2,...] :  Least common multiple  \r\n\
\n\
    n1, n2, ...\t64-bit unsigned integers  \n\
\n\
A maximum of 32 values are allowed.\n";


const wchar_t F_LPrimeHelpText[]	= L"@LPRIME[n] :  Leftward prime  \r\n\
\n\
    n\ta 64-bit unsigned integer \u2265 2  \n\
\n\
Returns the largest prime number not greater  \n\
than n.\n";


#ifdef PRIMES_RAND_UI64

const wchar_t _RandUI64HelpText[]	= L"_RANDUI64 :  64-bit random number  \r\n\
\n\
Returns a random number between 0 and 2^64 - 1.  \n";

#endif	//	PRIMES_RAND_UI64


struct _Factors {
	unsigned int n;
	unsigned __int64 f[MAX_UNIQUE_PRIME_FACTORS];
	unsigned int e[MAX_UNIQUE_PRIME_FACTORS];
};


#endif	//	PRIMES_MINIMAL



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



unsigned int Nonblanks( LPCTSTR String )
{
	if ( !String )
		return 0;
	

	const wchar_t *ch		= String;
	unsigned int k			= 0;


	while ( *ch ) {
		if ( !iswspace( *ch ) )
			k++;

		ch++;
	}

	return k;
}



/*

int DumpArray( unsigned __int64 *Values, unsigned int n )
{
	if ( !Values )
		return -1;


	for ( unsigned int i = 0; i < n; i++ )
		Printf( L"   Values[%u] = %I64u\n", i, Values[i] );

	CrLf();
	return 0;
}

*/



BOOL ValueInList( unsigned __int64 *Values, unsigned int n, unsigned __int64 Value )
{
	if ( !Values || !n )
		return FALSE;


	for ( unsigned int i = 0; i < n; i++ )
		if ( Values[i] == Value )
			return TRUE;

	return FALSE;
}



int EliminateRepeatedValues( unsigned __int64 *Values, unsigned int *n )
{
	if ( !Values )
		return -1;

	if ( *n < 2 )
		return 0;


	int k				= 0;


	//	DumpArray( Values, *n );

	for ( unsigned int i = 1; i < *n; i++ ) {
		if ( ValueInList( Values, i, Values[i] ) ) {
			for ( unsigned int j = i; j < ( *n - 1 ); j++ )
				Values[j] = Values[j + 1];

			*n -= 1;
			Values[*n] = 0;
			k++;
			i--;
		}
	}

	//	if ( k )
	//		DumpArray( Values, *n );

	return k;
}



#ifdef _WIN64

unsigned __int64 isqrt( unsigned __int64 n )
{
	if ( n < 2 )
		return n;


	unsigned __int64 s		= 0;
	unsigned __int64 t		= 0;
	DWORD i					= 0;
	

	if ( _BitScanReverse64( &i, n ) ) {
		s = (unsigned __int64) 1 << ( ( i + 2 ) / 2);
		t = ( s + ( n / s ) ) / 2;
	}
	else {
		s = n / 2;
		t = ( s + ( n / s ) ) / 2;
	}


	while ( t < s ) {
		s = t;
		t = ( s + ( n / s ) ) / 2;
	}

	return s;
}

#else	//	_WIN64

unsigned __int64 isqrt( unsigned __int64 n )
{
	if ( n < 2 )
		return n;


	unsigned __int64 s		= n / 2;
	unsigned __int64 t		= ( s + ( n / s ) ) / 2;
	

	while ( t < s ) {
		s = t;
		t = ( s + ( n / s ) ) / 2;
	}

	return s;
}

#endif	//	_WIN64


int IsPrime( unsigned __int64 Value )
{
	if ( Value < 2 )
		return 0;
	else if ( Value < 4 )
		return 1;
	else if ( ( Value & 1 ) == 0 )
		return 0;
	else if ( ( Value % 3 ) == 0 )
		return 0;


	unsigned __int64 div		= 5;
	unsigned __int64 limit		= isqrt( Value );


	while ( div <= limit ) {
		if ( ( Value % div ) == 0 )
			return 0;

		if ( ( Value % ( div + 2 ) ) == 0 )
			return 0;

		div += 6;
	}

	return 1;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int ParseInt64( LPTSTR inString, unsigned __int64 *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;


	unsigned __int64 Value	= 0;
	wchar_t *sep1			= NULL;


	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		Value = _wcstoui64( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = Value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//	parse decimal

		Value = _wcstoui64( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = Value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}

	return -1;
}



int LooksLikeANumber( LPCTSTR String )
{
	if ( !String )
		return 0;


	if ( ( String[0] == '0' ) && ( towupper( String[1] ) == 'X' ) && iswxdigit( String[2] ) ) {
		for ( const wchar_t *ch = String + 2; *ch; ch++ )
			if ( !iswxdigit( *ch ) )
				return 0;

		return 1;
	}
	else if ( iswdigit( String[0] ) ) {
		for ( const wchar_t *ch = String; *ch; ch++ )
			if ( !iswdigit( *ch ) )
				return 0;

		return 1;
	}

	return 0;
}



int EvaluateQuietly( LPTSTR String )
{
	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );
	int rv		= 0;


	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );		//  redirect stderr to nul:
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );		//  redirect stdout to nul:
	}
	
	rv = Evaluate( String );									//  and evaluate the expression

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );		//  put stderr back where we found it
		SetStdHandle( STD_OUTPUT_HANDLE , OldStdoutHandle );	//  put stdout back where we found it
		CloseHandle( NewStderrHandle );
	}

	return rv;
}



int Primes_ParseArg( LPCTSTR String, unsigned __int64 *outValue, int *outFormat )
{
	if ( !String || !outValue )
		return 1;


	wchar_t temp[MAX_ARG_LEN]		= L"";

	size_t len			= 0;
	int rv				= 0;


	wcscpy_s( temp, MAX_ARG_LEN, String );
	TrimLine( temp );

	len = wcslen( temp );
	if ( ( len >= 2 ) && outFormat ) {
		if ( ( temp[len - 2] == '=' ) && ( ( towupper( temp[len - 1] ) == 'X' ) || ( towupper( temp[len - 1] ) == 'H' ) ) ) {
			temp[len - 2] = '\0';
			TrimLine( temp );
			*outFormat = ( towupper( temp[len - 1] ) == 'H' ) ? PRIMES_HEX_ALT : PRIMES_HEX;
		}
	}

	if ( temp[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing argument\r\n" );
		return -1;
	}

	if ( !LooksLikeANumber( temp ) ) {
		if ( EvaluateQuietly( temp ) ) {
			CharUpperW( temp );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad expression: \"%s\"\r\n", temp );
			return -1;
		}
	}

	if ( rv = ParseInt64( temp, outValue, NULL ) ) {
		CharUpperW( temp );
		DisplayErrorHeader();

		if ( rv == 34 )
			Qprintf( ERRHANDLE, L"Value out of range: \"%s\"\r\n", temp );
		else
			Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", temp );

		return -1;
	}

	return 0;
}



DLLExports int WINAPI f_isprime( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned __int64 Value		= 0;
	int Format					= PRIMES_DECIMAL;
	int rv						= 0;


	if ( rv = Primes_ParseArg( String, &Value, &Format ) )
		return rv;

	_itow_s( IsPrime( Value ), String, MAX_ARG_LEN, 10 );
	return 0;
}



DLLExports int WINAPI f_prime( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned __int64 Value		= 0;
	int Format					= PRIMES_DECIMAL;
	int rv						= 0;


	if ( rv = Primes_ParseArg( String, &Value, &Format ) )
		return rv;

	if ( Value < 4 ) {
		Value = ( Value < 2 ) ? 2 : Value;
		goto Done;
	}


	if ( ( Value & 1 ) == 0 )
		Value++;


	do {
		if ( Value > LARGEST_PRIME ) {

#ifdef PRIMES_NO_OVERFLOW_ERROR

			switch ( Format ) {

			case PRIMES_HEX :
				wcscpy_s( String, MAX_ARG_LEN, L"0x1000000000000000D" );
				break;

			case PRIMES_HEX_ALT :
				wcscpy_s( String, MAX_ARG_LEN, L"1000000000000000D" );
				break;

			default :
				wcscpy_s( String, MAX_ARG_LEN, L"18446744073709551629" );
			}
			
			return 0;

#else	//	PRIMES_NO_OVERFLOW_ERROR

			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Integer overflow\r\n" );
			return -1;

#endif	//	PRIMES_NO_OVERFLOW_ERROR

		}

		if ( IsPrime( Value ) )
			break;

		Value += 2;
	}
	while ( TRUE );


Done:

	switch ( Format ) {

	case PRIMES_HEX :
		wsprintf( String, L"0x%I64X", Value );
		break;

	case PRIMES_HEX_ALT :
		wsprintf( String, L"%I64X", Value );
		break;

	default :
		_ui64tow_s( Value, String, MAX_ARG_LEN, 10 );
	}

	return 0;
}



#ifndef PRIMES_MINIMAL


DLLExports int WINAPI f_lprime( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned __int64 Value		= 0;
	int Format					= PRIMES_DECIMAL;
	int rv						= 0;


	if ( rv = Primes_ParseArg( String, &Value, &Format ) )
		return rv;

	if ( Value < 2 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Value must be at least 2: %I64u\r\n", Value );
		return -1;
	}
	else if ( Value < 4 )
		goto Done;


	if ( ( Value & 1 ) == 0 )
		Value--;

	while ( !IsPrime( Value ) )
		Value -= 2;


Done:

	switch ( Format ) {

	case PRIMES_HEX :
		wsprintf( String, L"0x%I64X", Value );
		break;

	case PRIMES_HEX_ALT :
		wsprintf( String, L"%I64X", Value );
		break;

	default :
		_ui64tow_s( Value, String, MAX_ARG_LEN, 10 );
	}

	return 0;
}



int Factorize( unsigned __int64 Value, _Factors *Factors )
{
	if ( !Factors || ( Value < 2 ) )
		return -1;


	unsigned __int64 div		= 5;
	unsigned int power			= 0;
	unsigned __int64 limit		= 0;
	unsigned int n				= 0;
	BOOL recalc					= FALSE;


	Factors->n = 0;


	power = 0;
	while ( ( Value & 1 ) == 0 ) {
		power++;
		Value >>= 1;
	}

	if ( power ) {
		Factors->f[n]	= 2;
		Factors->e[n++]	= power;
	}


	power = 0;
	while ( ( Value % 3 ) == 0 ) {
		power++;
		Value /= 3;
	}

	if ( power ) {
		Factors->f[n]	= 3;
		Factors->e[n++]	= power;
	}


	limit = isqrt( Value );


	while ( ( Value > 1 ) && ( div <= limit ) && ( div <= Value ) ) {

		power = 0;
		while ( ( Value % div ) == 0 ) {
			power++;
			Value /= div;
		}

		if ( power ) {
			Factors->f[n]	= div;
			Factors->e[n++]	= power;
			recalc			= TRUE;
		}

		power = 0;
		while ( ( Value % ( div + 2 ) ) == 0 ) {
			power++;
			Value /= ( div + 2 );
		}

		if ( power ) {
			Factors->f[n]	= div + 2;
			Factors->e[n++]	= power;
			recalc			= TRUE;
		}

		if ( recalc ) {
			limit = isqrt( Value );
			recalc = FALSE;
		}

		div += 6;
	}


	if ( Value > 1 ) {
		Factors->f[n]	= Value;
		Factors->e[n++]	= 1;
	}

	Factors->n = n;
	return n;
}



int FindFactors( unsigned __int64 Value, LPCTSTR MultOperator, LPCTSTR PowerOperator, int Format, LPTSTR String, size_t buflen )
{
	if ( ( Value < 2 ) || !String || ( buflen < 100 ) )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]		= L"";
	wchar_t DefaultMult[]			= L" \u00d7 ";

	_Factors Factors;
	size_t l			= 0;
	size_t m			= 0;
	int rv				= 0;


	String[0] = '\0';

	if ( !MultOperator )
		MultOperator = DefaultMult;
	else if ( MultOperator[0] == '\0' )
		MultOperator = DefaultMult;


	rv = Factorize( Value, &Factors );
	if ( rv < 0 )
		return rv;

	for ( unsigned int i = 0; i < Factors.n; i++ ) {

		if ( PowerOperator && ( Factors.e[i] > 1 ) ) {

			switch ( Format ) {

			case PRIMES_HEX :
				wsprintf( temp, L"0x%I64X%s0x%X%s", Factors.f[i], PowerOperator, Factors.e[i], MultOperator );
				break;

			case PRIMES_HEX_ALT :
				wsprintf( temp, L"%I64X%s%X%s", Factors.f[i], PowerOperator, Factors.e[i], MultOperator );
				break;

			default :
				wsprintf( temp, L"%I64u%s%u%s", Factors.f[i], PowerOperator, Factors.e[i], MultOperator );
			}

			wcscat_s( String, buflen, temp );
		}
		else {

			switch ( Format ) {

			case PRIMES_HEX :
				wsprintf( temp, L"0x%I64X%s", Factors.f[i], MultOperator );
				break;

			case PRIMES_HEX_ALT :
				wsprintf( temp, L"%I64X%s", Factors.f[i], MultOperator );
				break;

			default :
				wsprintf( temp, L"%I64u%s", Factors.f[i], MultOperator );
			}

			for ( unsigned int j = 0; j < Factors.e[i]; j++ )
				wcscat_s( String, buflen, temp );
		}
	}


	l = wcslen( String );
	m = wcslen( MultOperator );

	if ( l > m )
		String[l - m] = '\0';

	return 0;
}



#include "ParseArgs.cpp"

#define EXPAND_ESCAPES_REAL_BACKSPACE
#include "ExpandEscapes.cpp"



DLLExports int WINAPI f_factors( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t MultOperator[SHORT_BUF_LEN]		= L"";
	wchar_t PowerOperator[SHORT_BUF_LEN]	= L"";

	int Format					= PRIMES_DECIMAL;
	unsigned __int64 Value		= 0;
	int rv						= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_NO_TRIM );


	if ( arglen[0] ) {
		if ( rv = Primes_ParseArg( arg[0], &Value, &Format ) )
			return rv;

		if ( Value < 2 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Value must be at least 2: %I64u\r\n", Value );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing value\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		wcscpy_s( MultOperator, SHORT_BUF_LEN, arg[1] );
		ExpandEscapes( MultOperator, SHORT_BUF_LEN );
	}

	if ( arglen[2] ) {
		wcscpy_s( PowerOperator, SHORT_BUF_LEN, arg[2] );
		ExpandEscapes( PowerOperator, SHORT_BUF_LEN );
	}


	return FindFactors( Value, MultOperator[0] ? MultOperator : NULL, PowerOperator[0] ? PowerOperator : NULL, Format, String, MAX_ARG_LEN );
}



int Primes_GetFormat( LPTSTR String )
{
	if ( !String )
		return PRIMES_DECIMAL;


	size_t len			= wcslen( String );


	while ( len ) {
		if ( !iswspace( String[len - 1] ) )
			break;

		String[--len] = '\0';
	}


	if ( len >= 2 ) {
		if ( ( String[len - 2] == '=' ) && ( ( towupper( String[len - 1] ) == 'X' ) || ( towupper( String[len - 1] ) == 'H' ) ) ) {
			String[len - 2] = '\0';
			return ( towupper( String[len - 1] ) == 'H' ) ? PRIMES_HEX_ALT : PRIMES_HEX;
		}
	}

	return PRIMES_DECIMAL;
}



BOOL Primes_AllDivisible( unsigned __int64 *Values, unsigned int n, unsigned __int64 Divisor )
{
	for ( unsigned int i = 0; i < n; i++ )
		if ( Values[i] % Divisor )
			return FALSE;

	return TRUE;
}



unsigned __int64 Primes_TestDivisorPowers( unsigned __int64 *Values, unsigned int n, unsigned __int64 Divisor )
{
	unsigned __int64 LeastMult			= -1;
	unsigned __int64 temp				= 0;
	unsigned __int64 test				= 0;


	for ( unsigned int i = 0; i < n; i++ ) {
		temp = Values[i];
		test = 1;

		while ( ( temp % Divisor ) == 0 ) {
			temp /= Divisor;
			test *= Divisor;
		}

		if ( test == 1 )
			return 1;
		else if ( test < LeastMult )
			LeastMult = test;
	}

	return LeastMult;
}



unsigned __int64 Primes_GCF( unsigned __int64 *Values, unsigned int n )
{
	if ( !Values || !n )
		return 0;

	if ( n == 1 )
		return Values[0];


	_Factors Factors;

	unsigned __int64 min		= -1;
	unsigned __int64 max		= 0;
	unsigned __int64 gcf		= 1;


	for ( unsigned int i = 0; i < n; i++ ) {
		if ( Values[i] < min )
			min = Values[i];

		if ( Values[i] > max )
			max = Values[i];
	}

	if ( ( min == max ) || ( min < 2 ) )
		return min;

	if ( Primes_AllDivisible( Values, n, min ) )
		return min;


	Factorize( min, &Factors );

	for ( unsigned int i = 0; i < Factors.n; i++ )
		gcf *= Primes_TestDivisorPowers( Values, n, Factors.f[i] );

	return gcf;
}



DLLExports int WINAPI f_gcf( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned __int64 Values[MAX_VALUES];
	unsigned int n						= 0;
	int Format							= Primes_GetFormat( String );
	unsigned __int64 gcf				= 0;
	int rv								= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	for ( int i = 0; i < numargs; i++ ) {
		if ( Nonblanks( arg[i] ) ) {

			if ( n >= MAX_VALUES ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%u values maximum: arg%u = \"%s\"\r\n", MAX_VALUES, n + 1, arg[i] );
				return -1;
			}

#ifdef PRIMES_ALLOW_EXPRESSIONS_IN_LIST

			if ( rv = Primes_ParseArg( arg[i], Values + n, NULL ) )
				return rv;

#else	//	PRIMES_ALLOW_EXPRESSIONS_IN_LIST

			if ( ParseInt64( arg[i], Values + n, NULL ) ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad value for arg%u: \"%s\"\r\n", n + 1, arg[i] );
				return -1;
			}

#endif	//	PRIMES_ALLOW_EXPRESSIONS_IN_LIST

			if ( Values[n] == 0 ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Values must be nonzero: arg%u = \"%s\"\r\n", n + 1, arg[i] );
				return -1;
			}

			n++;
		}
	}


	if ( !n ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Must supply at least one value\r\n" );
		return -1;
	}


	EliminateRepeatedValues( Values, &n );

	gcf = Primes_GCF( Values, n );

	switch ( Format ) {

	case PRIMES_HEX :
		wsprintf( String, L"0x%I64X", gcf );
		break;

	case PRIMES_HEX_ALT :
		wsprintf( String, L"%I64X", gcf );
		break;

	default :
		_ui64tow_s( gcf, String, MAX_ARG_LEN, 10 );
	}

	return 0;
}



int Primes_LCM( unsigned __int64 *Values, unsigned int n, unsigned __int64 *Result )
{
	if ( !Values || !n || !Result )
		return -1;

	if ( n == 1 ) {
		*Result = Values[0];
		return 0;
	}


	_Factors *Factors				= NULL;
	unsigned __int64 *CommonF		= NULL;
	unsigned int *CommonE			= NULL;
	unsigned __int64 lcm			= 1;
	unsigned __int64 t				= 0;
	unsigned int tf					= 0;
	unsigned int cfc				= 0;
	unsigned int cfi				= -1;
	int rv							= -1;


	Factors = (_Factors*) malloc( n * sizeof( _Factors ) );
	if ( !Factors ) {
		rv = ERROR_NOT_ENOUGH_MEMORY;
		goto CleanupAndExit;
	}

	for ( unsigned int i = 0; i < n; i++ ) {
		Factors[i].n = 0;
		Factorize( Values[i], Factors + i );
		tf += Factors[i].n;
	}

	CommonF = (unsigned __int64*) malloc( tf * sizeof( unsigned __int64 ) );
	if ( !CommonF ) {
		rv = ERROR_NOT_ENOUGH_MEMORY;
		goto CleanupAndExit;
	}

	CommonE = (unsigned int*) malloc( tf * sizeof( unsigned int ) );
	if ( !CommonE ) {
		rv = ERROR_NOT_ENOUGH_MEMORY;
		goto CleanupAndExit;
	}


	for ( unsigned int i = 0; i < n; i++ ) {
		for ( unsigned int j = 0; j < Factors[i].n; j++ ) {
			cfi = UINT_MAX;
			for ( unsigned int q = 0; q < cfc; q++ )
				if ( Factors[i].f[j] == CommonF[q] )
					cfi = q;

			if ( cfi < UINT_MAX ) {
				if ( Factors[i].e[j] > CommonE[cfi] )
					CommonE[cfi] = Factors[i].e[j];
			}
			else {
				CommonF[cfc]	= Factors[i].f[j];
				CommonE[cfc]	= Factors[i].e[j];
				cfc++;
			}
		}
	}


	for ( unsigned int i = 0; i < cfc; i++ ) {
		for ( unsigned int j = 0; j < CommonE[i]; j++ ) {
			t = lcm;
			lcm *= CommonF[i];

			if ( ( lcm / CommonF[i] ) != t ) {
				rv = ERROR_ARITHMETIC_OVERFLOW;
				goto CleanupAndExit;
			}
		}
	}

	rv = 0;


CleanupAndExit:

	free( Factors );
	free( CommonF );
	free( CommonE );

	if ( !rv )
		*Result = lcm;

	return rv;
}



DLLExports int WINAPI f_lcm( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned __int64 Values[MAX_VALUES];
	unsigned int n						= 0;
	unsigned __int64 lcm				= 0;
	int Format							= Primes_GetFormat( String );
	int rv								= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	for ( int i = 0; i < numargs; i++ ) {
		if ( Nonblanks( arg[i] ) ) {

			if ( n >= MAX_VALUES ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"%u values maximum: arg%u = \"%s\"\r\n", MAX_VALUES, n + 1, arg[i] );
				return -1;
			}

#ifdef PRIMES_ALLOW_EXPRESSIONS_IN_LIST

			if ( rv = Primes_ParseArg( arg[i], Values + n, NULL ) )
				return rv;

#else	//	PRIMES_ALLOW_EXPRESSIONS_IN_LIST

			if ( ParseInt64( arg[i], Values + n, NULL ) ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad value: arg%u = \"%s\"\r\n", n + 1, arg[i] );
				return -1;
			}

#endif	//	PRIMES_ALLOW_EXPRESSIONS_IN_LIST

			if ( Values[n] == 0 ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Values must be nonzero: arg%u = \"%s\"\r\n", n + 1, arg[i] );
				return -1;
			}

			n++;
		}
	}


	if ( !n ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Must supply at least one value\r\n" );
		return -1;
	}


	EliminateRepeatedValues( Values, &n );

	if ( rv = Primes_LCM( Values, n, &lcm ) ) {
		DisplayErrorHeader();

		switch ( rv ) {

		case  ERROR_NOT_ENOUGH_MEMORY :
			Qprintf( ERRHANDLE, L"Not enough memory\r\n" );
			break;

		case ERROR_ARITHMETIC_OVERFLOW :
			Qprintf( ERRHANDLE, L"Integer overflow\r\n" );
			break;

		default :
			Qprintf( ERRHANDLE, L"Unknown error %d\r\n", rv );
		}

		return -1;
	}
	else {

		switch ( Format ) {

		case PRIMES_HEX :
			wsprintf( String, L"0x%I64X", lcm );
			break;

		case PRIMES_HEX_ALT :
			wsprintf( String, L"%I64X", lcm );
			break;

		default :
			_ui64tow_s( lcm, String, MAX_ARG_LEN, 10 );
		}

		return 0;
	}
}



DLLExports int WINAPI f_isqrt( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned __int64 Value		= 0;
	int Format					= PRIMES_DECIMAL;
	int rv						= 0;


	if ( rv = Primes_ParseArg( String, &Value, &Format ) )
		return rv;

	switch ( Format ) {

	case PRIMES_HEX :
		wsprintf( String, L"0x%I64X", isqrt( Value ) );
		break;

	case PRIMES_HEX_ALT :
		wsprintf( String, L"%I64X", isqrt( Value ) );
		break;

	default :
		_ui64tow_s( isqrt( Value ), String, MAX_ARG_LEN, 10 );
	}

	return 0;
}


#ifdef PRIMES_RAND_UI64


DLLExports int WINAPI _randui64( LPTSTR outString )
{
	if ( !outString )
		return 1;


	LARGE_INTEGER rv		= { 0, 0 };


	rand_s( (unsigned int*) &( rv.HighPart ) );
	rand_s( (unsigned int*) &( rv.LowPart ) );

	wsprintf( outString, L"%I64u", rv.QuadPart );
	return 0;
}


#endif	//	PRIMES_RAND_UI64

#endif	//	PRIMES_MINIMAL


BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{

#ifdef PRIMES_MINIMAL

	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/primes.html";
	piInfo.pszDescription	= (LPTSTR) L"Implements @ISPRIME and @PRIME.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

#else	//	PRIMES_MINIMAL

	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/primes.html";
	piInfo.pszDescription	= (LPTSTR) L"Implements @ISPRIME, @PRIME, @LPRIME, and related functions.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

#endif	//	PRIMES_MINIMAL

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



#ifdef PRIMES_MINIMAL

const int NumHelpTexts							= 2;
const wchar_t HelpFeature[NumHelpTexts][9]		= { L"@isprime", L"@prime" };
const wchar_t *HelpText[NumHelpTexts]			= { F_IsPrimeHelpText, F_PrimeHelpText };

#elif defined PRIMES_RAND_UI64

const int NumHelpTexts							= 8;
const wchar_t HelpFeature[NumHelpTexts][10]		= { L"@factors", L"@gcf", L"@isprime", L"@isqrt", L"@lcm", L"@lprime", L"@prime", L"_randui64" };
const wchar_t *HelpText[NumHelpTexts]			= { F_FactorsHelpText, F_GcfHelpText, F_IsPrimeHelpText, F_IsqrtHelpText, \
													F_LcmHelpText, F_LPrimeHelpText, F_PrimeHelpText, _RandUI64HelpText };
#else	//	PRIMES_MINIMAL

const int NumHelpTexts							= 7;
const wchar_t HelpFeature[NumHelpTexts][9]		= { L"@factors", L"@gcf", L"@isprime", L"@isqrt", L"@lcm", L"@lprime", L"@prime" };
const wchar_t *HelpText[NumHelpTexts]			= { F_FactorsHelpText, F_GcfHelpText, F_IsPrimeHelpText, F_IsqrtHelpText, \
													F_LcmHelpText, F_LPrimeHelpText, F_PrimeHelpText };
#endif	//	PRIMES_MINIMAL



DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}


/*

	0.99.0.0	2022-10-13

	0.99.1.0	2022-10-17
				if arg does not look like a number, try calling Evaluate()
				if arg ends in =X, return value in hexadecimal

	0.99.1.1	2022-10-18
				Renamed a preprocessor macro.  No functional changes.

	0.99.2.0	2022-10-21
				Diddled the out-of-range error message.

	0.99.3.0	2022-10-21
				Adds @FACTORS

	0.99.4.0	2022-10-22
				Adds @GCF

	0.99.5.0	2022-10-22
				Adds @ISQRT

	0.99.6.0	2022-10-23
				Adds a new common Factorize() routine for use by @FACTORS,
				@GCF, @LCM, etc.
				Adds @LCM
				Simplified @FACTORS

	0.99.7.0	2022-10-24
				Rewrote the overflow check in Primes_LCM().

	0.99.8.0	2022-10-24
				Expand character escapes in @FACTORS's multiplication
				and power symbols.

	0.99.9.0	2022-10-24
				Changed @GCF and @LCM to accept only integer values,
				not expressions.

	0.99.10.0	2022-10-27
				Added a compile-time option PRIMES_MINIMAL to remove
				all functions other than @ISPRIME and @PRIME.  No
				changes to functionality.

	0.99.11.0	2022-11-03
				Tweaks to the _FactStruct structure.

	0.99.12.0	2022-11-04
				Simplified the _Factorization struct, and dispensed
				with _FactStruct altogether.  Reworked Factorize() to
				recompute limit when possible: big performance boost!
				Added optional _RANDUI64 for testing purposes.

	0.99.13.0	2022-11-09
				Added =H for hexadecimal without the leading 0x.
				Enabled hex output in @FACTORS, God only knows why.

	0.99.14.0	2023-04-25
				Enabled hex output in @GCF and @LCM.  Fix for @LPRIME
				not formatting the output value correctly when the
				input value is less than 4.

	0.99.14.1	2023-10-12
				Updated the plugin's web address.

	0.99.15.0	2023-12-31
				Increased MAX_VALUES to 32.
				@GCF and @LCM now allow only one arg.  In this case,
				they simply return the value of the arg.
				@GCF and @LCM:  any arg parsing error will also report
				the number of the offending argument.
				Incorporated Vincent Fatica's suggestion to boost
				@ISQRT's performance (x64 build only; x86 works as
				before.)
				Uppercasing bad arguments in error messages is now
				done via CharUpperW() instead of _wcsupr_s().

	0.99.16.0	2024-01-01
				Now @GCF and @LCM ignore empty/blank arguments.
				Now @GCF and @LCM eliminate repeated values.

*/
